/*
 * $Id$
 */

/** <!--********************************************************************-->
 *
 * @defgroup ccwb Create C Wrapper Body traversal
 *
 * Module description goes here.
 *
 * @ingroup ccwb
 *
 * @{
 *
 *****************************************************************************/

/** <!--********************************************************************-->
 *
 * @file create_c_wrapper_body.c
 *
 * Prefix: CCWB
 *
 *****************************************************************************/
#include "create_c_wrapper_body.h"

#include "dbug.h"
#include "traverse.h"
#include "tree_basic.h"
#include "memory.h"
#include "filemgr.h"
#include "globals.h"
#include "bundle_to_fundef.h"
#include "build.h"
#include "tree_compound.h"

/** <!--********************************************************************-->
 *
 * @name INFO structure
 * @{
 *
 *****************************************************************************/
struct INFO {
    FILE *file;
};

/**
 * The info structure
 */
#define INFO_FILE(n) ((n)->file)

static info *
MakeInfo ()
{
    info *result;

    DBUG_ENTER ("MakeInfo");

    result = MEMmalloc (sizeof (info));

    INFO_FILE (result) = NULL;

    DBUG_RETURN (result);
}

static info *
FreeInfo (info *info)
{
    DBUG_ENTER ("FreeInfo");

    info = MEMfree (info);

    DBUG_RETURN (info);
}
/** <!--********************************************************************-->
 * @}  <!-- INFO structure -->
 *****************************************************************************/

/** <!--********************************************************************-->
 *
 * @name Static helper funcions
 * @{
 *
 *****************************************************************************/

/** <!--********************************************************************-->
 *
 * @fn void PrintFileHeader(FILE *file);
 *
 * @brief Prints the header for the file to be generated.
 *
 *****************************************************************************/
static void
PrintFileHeader (info *arg_info)
{
    DBUG_ENTER ("PrintFileHeader");

    fprintf (INFO_FILE (arg_info),
             "/*\n"
             " * C interface stub file.\n"
             " *\n"
             " * generated by sac4c %s (%s) rev %s\n"
             " */\n\n"
             "#include \"sacinterface.h\"\n"
             "#include \"header.h\"\n\n",
             global.version_id, build_style, build_rev);

    /*
     * descriptor alloc and free functions
     */
    fprintf (INFO_FILE (arg_info),
             "static SAC_array_descriptor_t makeScalarDesc() { \n"
             "  SAC_array_descriptor_t result \n"
             "    = (SAC_array_descriptor_t) SAC_MALLOC( BYTE_SIZE_OF_DESC( 0)); \n"
             "  DESC_DIM( result) = 0;\n"
             "  DESC_SIZE( result) = 1;\n"
             "  DESC_RC( result) = 1;\n"
             "  return( result);\n"
             "}\n\n");

    fprintf (INFO_FILE (arg_info),
             "static void freeScalarDesc( SAC_array_descriptor_t desc) { \n"
             "  SAC_FREE( desc);\n"
             "}\n\n");

    DBUG_VOID_RETURN;
}

/** <!--********************************************************************-->
 * @}  <!-- Static helper functions -->
 *****************************************************************************/

/** <!--********************************************************************-->
 *
 * @name Entry functions
 * @{
 *
 *****************************************************************************/
/** <!--********************************************************************-->
 *
 * @fn node *CCWBdoCreateCWrapperBody( node *syntax_tree)
 *
 *****************************************************************************/
node *
CCWBdoCreateCWrapperBody (node *syntax_tree)
{
    info *info;

    DBUG_ENTER ("CCWBdoCreateCWrapperBody");

    info = MakeInfo ();

    INFO_FILE (info) = FMGRwriteOpen ("%s/interface.c", global.tmp_dirname);

    PrintFileHeader (info);

    TRAVpush (TR_ccwb);
    syntax_tree = TRAVdo (syntax_tree, info);
    TRAVpop ();

    DBUG_PRINT ("TEMP", ("Template traversal complete."));

    INFO_FILE (info) = FMGRclose (INFO_FILE (info));

    info = FreeInfo (info);

    DBUG_RETURN (syntax_tree);
}

/** <!--********************************************************************-->
 * @}  <!-- Entry functions -->
 *****************************************************************************/

/** <!--********************************************************************-->
 *
 * @name Traversal functions
 * @{
 *
 *****************************************************************************/

/** <!--********************************************************************-->
 *
 * @fn node *CCWBfunbundle(node *arg_node, info *arg_info)
 *
 * @brief
 *
 *****************************************************************************/
node *
CCWBfunbundle (node *arg_node, info *arg_info)
{
    int noargs;
    int norets;
    int pos;

    DBUG_ENTER ("CCWBfunbundle");

    noargs = FUNBUNDLE_ARITY (arg_node);
    norets = TCcountRets (FUNDEF_RETS (FUNBUNDLE_FUNDEF (arg_node)));

    /*
     * fuunction header
     */
    fprintf (INFO_FILE (arg_info), "void %s(", FUNBUNDLE_EXTNAME (arg_node));

    for (pos = 0; pos < norets; pos++) {
        fprintf (INFO_FILE (arg_info), "void **ret%d%s", pos,
                 (norets - pos + noargs > 1) ? ", " : "");
    }
    for (pos = 0; pos < noargs; pos++) {
        fprintf (INFO_FILE (arg_info), "void *arg%d%s", pos,
                 (noargs - pos > 1) ? ", " : "");
    }

    fprintf (INFO_FILE (arg_info), ") {");

    /*
     * allocate arg descriptors and declare ret descriptors
     */
    for (pos = 0; pos < noargs; pos++) {
        fprintf (INFO_FILE (arg_info),
                 "SAC_array_descriptor_t arg%d_desc = makeScalarDesc();\n", pos);
    }
    for (pos = 0; pos < norets; pos++) {
        fprintf (INFO_FILE (arg_info), "SAC_array_descriptor_t ret%d_desc;\n", pos);
    }

    /*
     * call SAC fun
     */
    fprintf (INFO_FILE (arg_info), "%s%s( ", CWRAPPER_PREFIX,
             FUNBUNDLE_EXTNAME (arg_node));
    for (pos = 0; pos < norets; pos++) {
        fprintf (INFO_FILE (arg_info), "ret%d, &ret%d_desc%s", pos, pos,
                 (norets - pos + noargs > 1) ? ", " : "");
    }
    for (pos = 0; pos < noargs; pos++) {
        fprintf (INFO_FILE (arg_info), "arg%d, arg%d_desc%s", pos, pos,
                 (noargs - pos > 1) ? ", " : "");
    }
    fprintf (INFO_FILE (arg_info), ");\n");

    /*
     * free return descs
     */
    for (pos = 0; pos < norets; pos++) {
        fprintf (INFO_FILE (arg_info), "freeScalarDesc( ret%d_desc);\n", pos);
    }

    fprintf (INFO_FILE (arg_info), "}\n\n");

    if (FUNBUNDLE_NEXT (arg_node) != NULL) {
        FUNBUNDLE_NEXT (arg_node) = TRAVdo (FUNBUNDLE_NEXT (arg_node), arg_info);
    }

    DBUG_RETURN (arg_node);
}

/** <!--********************************************************************-->
 * @}  <!-- Traversal functions -->
 *****************************************************************************/

/** <!--********************************************************************-->
 * @}  <!-- Create C Wrapper Body template -->
 *****************************************************************************/
