/*
 * $Id$
 */

#include "dependencies.h"

#include <libgen.h>

#include "globals.h"
#include "stringset.h"
#include "tree_basic.h"
#include "modulemanager.h"
#include "filemgr.h"
#include "str_buffer.h"
#include "str.h"
#include "memory.h"
#include "namespaces.h"
#include "ctinfo.h"
#include "dbug.h"

/*
 * functions for generating the dependency table c file
 */
static void
GenerateDependencyTableHead (FILE *file)
{
    DBUG_ENTER ("GenerateDependencyTableHead");

    fprintf (file, "/* generated by sac2c %s */\n\n", global.version_id);
    fprintf (file, "#include \"sac_serialize.h\"\n\n");
    fprintf (file, "void *__%s__DEPTAB() {\n", global.modulename);
    fprintf (file, "void *result = (void *) 0;\n");

    DBUG_VOID_RETURN;
}

static void
GenerateDependencyTableTail (FILE *file)
{
    DBUG_ENTER ("GenerateDependencyTableTail");

    fprintf (file, "return(result);\n}\n");

    DBUG_VOID_RETURN;
}

static void *
TableEntriesFoldFun (const char *val, strstype_t kind, void *rest)
{
    str_buf *result = (str_buf *)rest;

    DBUG_ENTER ("TableEntriesFoldFun");

    switch (kind) {
    case STRS_saclib:
    case STRS_extlib:
        result = SBUFprintf ((str_buf *)rest, "result = STRSadd( \"%s\", %d, result);\n",
                             val, kind);
        break;
    default:
        break;
    }

    DBUG_RETURN (result);
}

static void
GenerateDependencyTableEntries (stringset_t *deps, FILE *file)
{
    str_buf *buffer;
    char *string;

    DBUG_ENTER ("GenerateDependencyTableEntries");

    buffer = SBUFcreate (4096);

    buffer = STRSfold (&TableEntriesFoldFun, deps, buffer);

    string = SBUF2str (buffer);

    fprintf (file, string);

    string = MEMfree (string);
    buffer = SBUFfree (buffer);

    DBUG_VOID_RETURN;
}

static void
GenerateDependencyTable (stringset_t *deps)
{
    FILE *file;

    DBUG_ENTER ("GenerateDependencyTable");

    file = FMGRwriteOpen ("%s/dependencytable.c", global.tmp_dirname);

    GenerateDependencyTableHead (file);

    GenerateDependencyTableEntries (deps, file);

    GenerateDependencyTableTail (file);

    fclose (file);

    DBUG_VOID_RETURN;
}

/*
 * functions for printing dependency information
 */
static void *
BuildDepClosFoldFun (const char *entry, strstype_t kind, void *rest)
{
    stringset_t *result = NULL;
    module_t *module;

    DBUG_ENTER ("BuildDepClosFoldFun");

    DBUG_PRINT ("DEP", ("Adding dependencies for %s...", entry));

    if (kind == STRS_saclib) {
        module = MODMloadModule (entry);
        result = MODMgetDependencyTable (module);
        module = MODMunLoadModule (module);
    }

    result = STRSjoin ((stringset_t *)rest, result);

    DBUG_RETURN ((void *)result);
}

static stringset_t *
BuildDependencyClosure (stringset_t *deps)
{
    stringset_t *result;

    DBUG_ENTER ("BuildDependencyClosure");

    result = STRSfold (&BuildDepClosFoldFun, deps, NULL);

    if (result != NULL) {
        /**
         * we have found some second level dependencies,
         * so next we build the closure over them.
         */
        result = BuildDependencyClosure (result);
    }

    result = STRSjoin (result, deps);

    DBUG_RETURN (result);
}

static void *
PrintLibDepFoldFun (const char *entry, strstype_t kind, void *modname)
{
    DBUG_ENTER ("PrintLibDepFoldFun");

    /*
     * if the dependency is a sac library, we print a make
     * rule for the given library
     */
    if (kind == STRS_saclib) {
        char *libname;
        char *libfile;

        libname = MEMmalloc (sizeof (char) * (STRlen (entry) + 5));
        sprintf (libname, "%s.sac", entry);

        libfile = STRcpy (FMGRfindFile (PK_imp_path, libname));
        libname = MEMfree (libname);

        if (libfile != NULL) {
            char *libdir = dirname (libfile);

            printf ("alldeps_%s: \\\n  %s\n\n", (char *)modname, entry);

            printf (".PHONY: \\\n  %s\n\n", entry);

            /**
             *
             * The space before the colon is essential here!
             * This allows us to easily distinguish this case from the
             * alldeps rules, in particular, the empty rule in
             * doPrintLibDependencies
             */
            printf ("%s :\n\t( cd %s; $(MAKE) lib%sTree.so)\n\n", entry, libdir, entry);
        }

        libfile = MEMfree (libfile);
    }

    DBUG_RETURN (modname);
}

static void
doPrintLibDependencies (node *tree)
{
    DBUG_ENTER ("doPrintLibDependencies");

#ifndef DBUG_OFF
    printf ("#\n# extended dependencies for file %s\n#\n\n", global.filename);
#endif

    /**
     * First, we give an empty rule. This is required iff there
     * are no dependencies!
     * The non-space before the colon is essential here!
     * Cf. PrintLibDepFoldFun!
     */
    printf ("alldeps_%s:\n", NSgetName (MODULE_NAMESPACE (tree)));

    STRSfold (&PrintLibDepFoldFun, global.dependencies,
              (void *)NSgetName (MODULE_NAMESPACE (tree)));

    DBUG_VOID_RETURN;
}

static void
PrintSACLib (const char *name)
{
    char *filename;
    char *result;

    DBUG_ENTER ("PrintSACLib");

    /*
     * first try to find the .so file
     */

    filename = MEMmalloc (sizeof (char) * (STRlen (name) + 11));
    sprintf (filename, "lib%sTree.so", name);

    result = STRcpy (FMGRfindFile (PK_lib_path, filename));

    filename = MEMfree (filename);

    if (result == NULL) {
        /*
         * now try to find the .sac file
         */

        filename = MEMmalloc (sizeof (char) * (STRlen (name) + 5));
        sprintf (filename, "%s.sac", name);

        result = STRcpy (FMGRfindFile (PK_imp_path, filename));

        filename = MEMfree (filename);
    }

    if (result == NULL) {
        /*
         * otherwise use the pure filename
         */

        result = MEMmalloc (sizeof (char) * (STRlen (name) + 11));
        sprintf (result, "lib%sTree.so", name);
    }

    printf (" \\\n  %s", result);

    result = MEMfree (result);

    DBUG_VOID_RETURN;
}

static void
PrintObjFile (const char *name)
{
    DBUG_ENTER ("PrintObjFile");

    printf (" \\\n  %s", name);

    DBUG_VOID_RETURN;
}

static void *
PrintDepFoldFun (const char *entry, strstype_t kind, void *rest)
{
    DBUG_ENTER ("PrintDepFoldFun");

    switch (kind) {
    case STRS_saclib:
        PrintSACLib (entry);
        break;
    case STRS_objfile:
        PrintObjFile (entry);
        break;
    default:
        break;
    }

    DBUG_RETURN ((void *)NULL);
}

static void
PrintTargetName (node *tree)
{
    DBUG_ENTER ("PrintTargetName");

    switch (MODULE_FILETYPE (tree)) {
    case F_prog:
        printf ("%s:", global.outfilename);
        break;
    case F_modimp:
    case F_classimp:
        printf ("%slib%sTree.so %slib%sMod.a %slib%sMod.so:", global.targetdir,
                NSgetName (MODULE_NAMESPACE (tree)), global.targetdir,
                NSgetName (MODULE_NAMESPACE (tree)), global.targetdir,
                NSgetName (MODULE_NAMESPACE (tree)));
        break;
    default:
        DBUG_ASSERT (0, ("unknown file type found!"));
        break;
    }

    DBUG_VOID_RETURN;
}

node *
DEPdoPrintDependencies (node *syntax_tree)
{
    DBUG_ENTER ("DEPdoPrintDependencies");

#ifndef DBUG_OFF
    printf ("#\n# dependencies for file %s\n#\n\n", global.filename);
#endif

    /*
     * first, print how the output will be named
     */
    PrintTargetName (syntax_tree);

    /*
     * now add the dependencies
     */
    STRSfold (&PrintDepFoldFun, global.dependencies, NULL);

    /*
     * and finally two newline to make it look nicer
     */
    printf ("\n\n");

    /*
     * if we are in special Mlib (aka sbs) mode, we have to print
     * some more dependencies, handeled by doPrintLibDependencies
     */
    if (global.makelibdeps) {
        doPrintLibDependencies (syntax_tree);
    }

    CTIterminateCompilation (syntax_tree);

    DBUG_RETURN (syntax_tree);
}

/*
 * We use a rather weird signature here to fit the framework of the
 * compiler phase mechanism.
 */
node *
DEPdoHandleDependencies (node *syntax_tree)
{
    DBUG_ENTER ("DEPdoHandleDependencies");

    if ((global.filetype == F_modimp) || (global.filetype == F_classimp)) {
        /*
         * finally generate the dependency table.
         * we do this here as new dependencies may be introduced
         * during the compilation steps up to here
         */
        GenerateDependencyTable (global.dependencies);
    } else {
        /*
         * for programs, we build the closure of all dependencies.
         * again, we cannot do this earlier, as new dependencies
         * might have been introduced until here.
         */
        global.dependencies = BuildDependencyClosure (global.dependencies);
    }

    DBUG_RETURN (syntax_tree);
}
